/*-------------------------------------------------------------------------------
	DSerial Simple Console
	(C) 2007 by Alexei Karpenko

	A quick keyboard hack.

	Keyboard graphics by StoneCypher http://sc.tri-bit.com/Keyboard
-------------------------------------------------------------------------------*/

#include <nds.h>
#include <stdio.h>
#include <string.h>
#include "keyboard.h"
#include "keyboard_map.h"
#include "keyboard_pal.h"
#include "keyboard_raw.h"

uint16 * KeyboardMap = (uint16 *) BG_MAP_RAM_SUB(31);
uint16 * Keyboard[2][2];
bool     KeyboardShift = false;
bool	 KeyboardCapslock = false;

typedef struct {
	int x;
	int y;
} Point2D;

Point2D KeyboardFrom;
void (*keyboardKeycodeHandler)(int keycode);

static const int KeyboardMeta[5][32][4] = { /* tuples of (x position, length, normal char, shifted char) for each tile */

	{ {0,2,DVK_FOLD,DVK_FOLD} , {0,2,DVK_FOLD,DVK_FOLD}, {2,0,NOKEY,NOKEY}, {3,2,'1','!'}, {3,2,'1','!'}, {5,2,'2','@'}, {5,2,'2','@'}, {7,2,'3','#'}, {7,2,'3','#'}, {9,2,'4','$'}, {9,2,'4','$'}, {11,2,'5','%'}, {11,2,'5','%'}, {13,2,'6','^'}, {13,2,'6','^'}, {15,2,'7','&'}, 
      {15,2,'7','&'}, {17,2,'8','*'}, {17,2,'8','*'}, {19,2,'9','('}, {19,2,'9','('}, {21,2,'0',')'}, {21,2,'0',')'}, {23,2,'-','_'}, {23,2,'-','_'}, {25,2,'=','+'}, {25,2,'=','+'}, {27,5,DVK_BACKSPACE,DVK_BACKSPACE}, {27,5,DVK_BACKSPACE,DVK_BACKSPACE}, {27,5,DVK_BACKSPACE,DVK_BACKSPACE}, {27,5,DVK_BACKSPACE,DVK_BACKSPACE}, {27,5,DVK_BACKSPACE,DVK_BACKSPACE} },

    { {0,4,DVK_TAB,DVK_TAB}, {0,4,DVK_TAB,DVK_TAB}, {0,4,DVK_TAB,DVK_TAB}, {0,4,DVK_TAB,DVK_TAB}, {4,2,'q','Q'}, {4,2,'q','Q'}, {6,2,'w','W'}, {6,2,'w','W'}, {8,2,'e','E'}, {8,2,'e','E'}, {10,2,'r','R'}, {10,2,'r','R'}, {12,2,'t','T'}, {12,2,'t','T'}, {14,2,'y','Y'}, {14,2,'y','Y'}, 
	{16,2,'u','U'}, {16,2,'u','U'}, {18,2,'i','I'}, {18,2,'i','I'}, {20,2,'o','O'}, {20,2,'o','O'}, {22,2,'p','P'}, {22,2,'p','P'}, {24,2,'[','{'}, {24,2,'[','{'}, {26,2,']','}'}, {26,2,']','}'}, {28,2,'\\','|'}, {28,2,'\\','|'}, {30,2,'`','~'}, {30,2,'`','~'} },

    { {0,5,DVK_CAPS,DVK_CAPS}, {0,5,DVK_CAPS,DVK_CAPS}, {0,5,DVK_CAPS,DVK_CAPS}, {0,5,DVK_CAPS,DVK_CAPS}, {0,5,DVK_CAPS,DVK_CAPS}, {5,2,'a','A'}, {5,2,'a','A'}, {7,2,'s','S'}, {7,2,'s','S'}, {9,2,'d','D'}, {9,2,'d','D'}, {11,2,'f','F'}, {11,2,'f','F'}, {13,2,'g','G'}, {13,2,'g','G'}, {15,2,'h','H'}, 
	{15,2,'h','H'}, {17,2,'j','J'}, {17,2,'j','J'}, {19,2,'k','K'}, {19,2,'k','K'}, {21,2,'l','L'}, {21,2,'l','L'}, {23,2,';',':'}, {23,2,';',':'}, {25,2,'\'','"'}, {25,2,'\'','"'}, {27,5,DVK_ENTER,DVK_ENTER}, {27,5,DVK_ENTER,DVK_ENTER}, {27,5,DVK_ENTER,DVK_ENTER}, {27,5,DVK_ENTER,DVK_ENTER}, {27,5,DVK_ENTER,DVK_ENTER} },

    { {0,6,DVK_SHIFT,DVK_SHIFT}, {0,6,DVK_SHIFT,DVK_SHIFT}, {0,6,DVK_SHIFT,DVK_SHIFT}, {0,6,DVK_SHIFT,DVK_SHIFT}, {0,6,DVK_SHIFT,DVK_SHIFT}, {0,6,DVK_SHIFT,DVK_SHIFT}, {6,2,'z','Z'}, {6,2,'z','Z'}, {8,2,'x','X'}, {8,2,'x','X'}, {10,2,'c','C'}, {10,2,'c','C'}, {12,2,'v','V'}, {12,2,'v','V'}, {14,2,'b','B'}, {14,2,'b','B'}, 
      {16,2,'n','N'}, {16,2,'n','N'}, {18,2,'m','M'}, {18,2,'m','M'}, {20,2,',','<'}, {20,2,',','<'}, {22,2,'.','>'}, {22,2,'.','>'}, {24,2,'/','?'}, {24,2,'/','?'}, {26,0,NOKEY,NOKEY}, {26,0,NOKEY,NOKEY}, {28,2,DVK_UP,DVK_UP}, {28,2,DVK_UP,DVK_UP}, {30,0,NOKEY,NOKEY}, {30,0,NOKEY,NOKEY} },

    { {0,5,DVK_CTRL,DVK_CTRL}, {0,5,DVK_CTRL,DVK_CTRL}, {0,5,DVK_CTRL,DVK_CTRL}, {0,5,DVK_CTRL,DVK_CTRL}, {0,5,DVK_CTRL,DVK_CTRL}, {5,4,DVK_ALT,DVK_ALT}, {5,4,DVK_ALT,DVK_ALT}, {5,4,DVK_ALT,DVK_ALT}, 
      {5,4,DVK_ALT,DVK_ALT}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, 
      {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {9,12,DVK_SPACE,DVK_SPACE}, {21,5,DVK_MENU,DVK_MENU}, {21,5,DVK_MENU,DVK_MENU}, {21,5,DVK_MENU,DVK_MENU}, 
      {21,5,DVK_MENU,DVK_MENU}, {21,5,DVK_MENU,DVK_MENU}, {26,2,DVK_LEFT,DVK_LEFT}, {26,2,DVK_LEFT,DVK_LEFT}, {28,2,DVK_DOWN,DVK_DOWN}, {28,2,DVK_DOWN,DVK_DOWN}, {30,2,DVK_RIGHT,DVK_RIGHT}, {30,2,DVK_RIGHT,DVK_RIGHT} }
};

/*-------------------------------------------------------------------------------*/
void keyboardDefaultKeycodeHandler(int keycode) {
/*-------------------------------------------------------------------------------*/
	if (keycode > 0) {
		iprintf("%c", (char) keycode);
	}
}

/*-------------------------------------------------------------------------------*/
void keyboardInit() {
/*-------------------------------------------------------------------------------*/
	int i;

	Keyboard[0][0] = ((uint16 *) keyboard_map) + 320;	// normal
	Keyboard[1][0] = (uint16 *) keyboard_map;			// shifted
	Keyboard[0][1] = ((uint16 *) keyboard_map) + 960;	// normal + pressed
	Keyboard[1][1] = ((uint16 *) keyboard_map) + 640;	// shifted + pressed

	keyboardKeycodeHandler = keyboardDefaultKeycodeHandler;

	// copy bitmap
	swiFastCopy(keyboard_raw, (u16 *) BG_BMP_RAM_SUB(0), COPY_MODE_HWORD | keyboard_raw_size/2);
	// copy map
	for(i = 0; i < 32 * 32; i++)
		KeyboardMap[i] = 2;
	swiFastCopy(Keyboard[0][0], KeyboardMap, COPY_MODE_HWORD | 320/2); 
	// copy palette
	swiFastCopy(keyboard_pal, (u16 *) BG_PALETTE_SUB, COPY_MODE_HWORD | keyboard_pal_size/2); 
}

/*-------------------------------------------------------------------------------*/
void keyboardShift(bool shift) {
/*-------------------------------------------------------------------------------*/
	KeyboardShift = shift;
	swiFastCopy(Keyboard[shift?1:0][0], KeyboardMap, COPY_MODE_HWORD | 320/2);
}

/*-------------------------------------------------------------------------------*/
void keyboardState(int r, int c, bool pressed) {
/*-------------------------------------------------------------------------------*/
	int i;
	int r2 = r/2;

	for (i = 0; i < KeyboardMeta[r2][c][1]; i++) {
		int tilex = KeyboardMeta[r2][c][0] + i;
		KeyboardMap[32*r2*2+tilex] = Keyboard[KeyboardShift?1:0][pressed?1:0][32*r2*2+tilex];
		KeyboardMap[32*(r2*2+1)+tilex] = Keyboard[KeyboardShift?1:0][pressed?1:0][32*(r2*2+1)+tilex];
	}
}

/*-------------------------------------------------------------------------------*/
void keyboardPress(int r, int c, bool repeat) {
/*-------------------------------------------------------------------------------*/
	int keycode = KeyboardMeta[r/2][c][KeyboardShift?3:2];
	if (keycode > 0) {
		if (KeyboardShift && !KeyboardCapslock) {
			keyboardShift(false);
		}
		keyboardState(r, c, true);
	} else if (keycode == DVK_SHIFT && !repeat) {
		KeyboardCapslock = false;
		keyboardShift(true);
		keyboardState(r, c, true);
	} else if (keycode == DVK_CAPS && !repeat) {
		KeyboardCapslock = !KeyboardCapslock;
		keyboardShift(KeyboardCapslock);
		keyboardState(r, c, KeyboardCapslock);
	}

	if (keyboardKeycodeHandler != NULL) {
		if (keycode > 0 || !repeat) {
			keyboardKeycodeHandler(keycode);
		}
	}
}

/*-------------------------------------------------------------------------------*/
void keyboardCancel(int r, int c) {
/*-------------------------------------------------------------------------------*/
	int keycode = KeyboardMeta[r/2][c][KeyboardShift?3:2];
	if (keycode == DVK_SHIFT || keycode == DVK_CAPS) {
		keyboardShift(false);
		KeyboardCapslock = false;
	}

	keyboardState(r, c, false);
}

/*-------------------------------------------------------------------------------*/
void keyboardRelease(int r, int c) {
/*-------------------------------------------------------------------------------*/
	int keycode = KeyboardMeta[r/2][c][KeyboardShift?3:2];

	if (keycode != DVK_SHIFT && keycode != DVK_CAPS) {
		keyboardState(r, c, false);
	}
}

/*-------------------------------------------------------------------------------*/
void keyboardUpdate(touchPosition pos) {
/*-------------------------------------------------------------------------------*/
	uint16 tilex = pos.px / 8;
	uint16 tiley = pos.py / 8;

	if (keysDown() & KEY_TOUCH) {
		if (tiley < 10) {
			KeyboardFrom.x = tilex;
			KeyboardFrom.y = tiley;
			keyboardPress(tiley, tilex , false);
			keysSetRepeat(40,20);
		} else {
			KeyboardFrom.x = -1;
		}

	} else if (keysDownRepeat() & KEY_TOUCH) {
		if (KeyboardFrom.x != -1) {
			keyboardPress(KeyboardFrom.y, KeyboardFrom.x, true);
		}
	} else if (keysUp() & KEY_TOUCH) {
		if (KeyboardFrom.x != -1) {
			if (KeyboardMeta[tiley/2][tilex][2] == KeyboardMeta[KeyboardFrom.y/2][KeyboardFrom.x][2]) {
				keyboardRelease(KeyboardFrom.y, KeyboardFrom.x);
			} else {
				keyboardCancel(KeyboardFrom.y, KeyboardFrom.x);
			}
		}
	}
}

/*-------------------------------------------------------------------------------*/
void keyboardSetKeycodeHandler(void (*keycodeHandler)(int keycode)) {
/*-------------------------------------------------------------------------------*/
	keyboardKeycodeHandler = keycodeHandler;
}
