/*-------------------------------------------------------------------------------
	DSerial Simple Console
	(C) 2007 by Alexei Karpenko
-------------------------------------------------------------------------------*/

#include <nds.h>
#include <stdio.h>
#include <string.h>
#include "dserial.h"
#include "firmware_bin.h"
#include "keyboard.h"

/*-------------------------------------------------------------------------------*/
void keycodeHandler(int keycode) {
/*-------------------------------------------------------------------------------*/
	if (keycode > 0) {
		char c = (char) keycode;
		iprintf("%c", c);				/* local echo */
		dseUartSendBuffer(&c, 1, true); /* blocking uart */
	}
}

/*-------------------------------------------------------------------------------*/
int main(void) {
/*-------------------------------------------------------------------------------*/

	videoSetMode(MODE_0_2D | DISPLAY_BG0_ACTIVE);
	videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankC(VRAM_C_SUB_BG);

	BG0_CR = BG_MAP_BASE(31);
	SUB_BG0_CR = BG_COLOR_256 | BG_MAP_BASE(31);

	BG_PALETTE[255] = RGB15(31,31,31);	//by default font will be rendered with color 255

	consoleInitDefault((u16*)SCREEN_BASE_BLOCK(31), (u16*)CHAR_BASE_BLOCK(0), 16);

	keyboardInit();
	keyboardSetKeycodeHandler(keycodeHandler);

	irqInit();
	irqSet(IRQ_VBLANK, 0);





	iprintf("-- DSerial Console --\n\n");

	while(!dseInit()) {
		iprintf("DSerial is not inserted!\n");
		iprintf("Insert DSerial and press A.\n");
		
		while(~keysDown() & KEY_A) { swiWaitForVBlank(); scanKeys(); }
		while(~keysUp() & KEY_A) { swiWaitForVBlank(); scanKeys(); }
	}
	iprintf("DSerial initialized!\n");

	/* Now see if the firmware needs to be updated and update if necessary */

	if (dseMatchFirmware((char *) firmware_bin, firmware_bin_size)) {
		iprintf("Firmwares match!\n");
	} else {
		iprintf("Updating firmware...\n");
		dseUploadFirmware((char *) firmware_bin, firmware_bin_size);
		iprintf("Done!\n");
	}
	dseBoot();
	swiDelay(9999); /* Wait for the FW to boot */
	if (dseStatus() == FIRMWARE) {
		iprintf("Firmware booted successfully!\n");
	} else {
		iprintf("Failed to boot firmware!\n");
	}

	/*  Set DSerial modes

		DSerial can interface with RS232 and TTL/CMOS level UART ports.
		If you need RS232, then set ENABLE_RS232 and wire up to RS232 connector.
		If you need CMOS/TTL, then set ENABLE_CMOS and wire up to IO2 connector.

		See http://www.beyondlogic.org/serial/serial1.htm#40 for explanation of
		the difference between RS232 and CMOS/TTL levels.
	*/
	dseSetModes(ENABLE_RS232 /* | ENABLE_SERVO */);

	/* Now DSerial is ready to be used */

	/*	You can set your own UART receive handler that will be called on an
		interrupt.

		See dseUartDefaultReceiveHandler() in dserial.c for an example.
	*/
	/* dseUartSetReceiveHandler(yourOwnUartHandler); */





	iprintf("\n");
	iprintf("A - Sends UART test string\n");
	iprintf("B - N/A\n");
	iprintf("X - Sets UART baud rate to 4800\n");
	iprintf("Y - Reads tilt data\n");
	iprintf("L - N/A\n");
	iprintf("R - Bootloader <--> Firmware\n");
	iprintf("\n");

	while(1) {
		touchPosition pos;
	
		swiWaitForVBlank();

		pos = touchReadXY();
		scanKeys();
		keyboardUpdate(pos);

		if(keysDown() & KEY_A) {
			dseUartSendBuffer("Hello!", 6, true); /* blocking UART send */
		} else if(keysDown() & KEY_B) {
			iprintf("<N/A>");
		} else if(keysDown() & KEY_X) {
			dseUartSetBaudrate(4800); /* at boot-up, baudrate is set to 115200 bps */
		} else if(keysDown() & KEY_Y) {;
			char buffer[32];
			uint8 size = dseAdcRead(buffer);
			if (size >= 2) {
				iprintf("Tilt = (0x%02X, 0x%02X)\n", buffer[0], buffer[1]);
			}
		} else if(keysDown() & KEY_R) {
			iprintf("<Boot>");
			dseBoot();
		} else if(keysDown() & KEY_L) {
			iprintf("<N/A>");
		}

		if(keysHeld() & KEY_TOUCH) {	/* R/C servo test code */
			if (pos.py >= 80) {			/* touch below keyboard */
				if (pos.y < 80 + 37) {
					iprintf("S[0] = %i\n", pos.px);
					dseServoSet(0, pos.px); /* set servo position to the x coordinate */
				} else if (pos.y < 80 + 2*37) {
					iprintf("S[1] = %i\n", pos.px);
					dseServoSet(1, pos.px); /* set servo position to the x coordinate */
				} else {
					iprintf("S[2] = %i\n", pos.px);
					dseServoSet(2, pos.px); /* set servo position to the x coordinate */
				}
			}
		}
	}

	return 0;
}
